/*++
/* NAME
/*	smtpd_sasl_glue 3
/* SUMMARY
/*	Postfix SMTP server, SASL support interface
/* SYNOPSIS
/*	#include "smtpd_sasl_glue.h"
/*
/*	void	smtpd_sasl_initialize()
/*
/*	void	smtpd_sasl_connect(state, sasl_opts_name, sasl_opts_val)
/*	SMTPD_STATE *state;
/*
/*	char	*smtpd_sasl_authenticate(state, sasl_method, init_response)
/*	SMTPD_STATE *state;
/*	const char *sasl_method;
/*	const char *init_response;
/*
/*	void	smtpd_sasl_logout(state)
/*	SMTPD_STATE *state;
/*
/*	void	smtpd_sasl_disconnect(state)
/*	SMTPD_STATE *state;
/* DESCRIPTION
/*	This module encapsulates most of the detail specific to SASL
/*	authentication.
/*
/*	smtpd_sasl_initialize() initializes the SASL library. This
/*	routine should be called once at process start-up. It may
/*	need access to the file system for run-time loading of
/*	plug-in modules. There is no corresponding cleanup routine.
/*
/*	smtpd_sasl_connect() performs per-connection initialization.
/*	This routine should be called once at the start of every
/*	connection. The sasl_opts_name and sasl_opts_val parameters
/*	are the postfix configuration parameters setting the security
/*	policy of the SASL authentication.
/*
/*	smtpd_sasl_authenticate() implements the authentication dialog.
/*	The result is a null pointer in case of success, an SMTP reply
/*	in case of failure. smtpd_sasl_authenticate() updates the
/*	following state structure members:
/* .IP sasl_method
/*	The authentication method that was successfully applied.
/*	This member is a null pointer in the absence of successful
/*	authentication.
/* .IP sasl_username
/*	The username that was successfully authenticated.
/*	This member is a null pointer in the absence of successful
/*	authentication.
/* .PP
/*	smtpd_sasl_logout() cleans up after smtpd_sasl_authenticate().
/*	This routine exists for the sake of symmetry.
/*
/*	smtpd_sasl_disconnect() performs per-connection cleanup.
/*	This routine should be called at the end of every connection.
/*
/*	Arguments:
/* .IP state
/*	SMTP session context.
/* .IP sasl_method
/*	A SASL mechanism name
/* .IP init_reply
/*	An optional initial client response.
/* DIAGNOSTICS
/*	All errors are fatal.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Initial implementation by:
/*	Till Franke
/*	SuSE Rhein/Main AG
/*	65760 Eschborn, Germany
/*
/*	Adopted by:
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <stdlib.h>
#include <string.h>

/* Utility library. */

#include <base64_code.h>
#include <msg.h>
#include <mymalloc.h>
#include <namadr_list.h>
#include <name_mask.h>

/* Global library. */

#include <mail_params.h>
#include <smtp_stream.h>

#ifdef USE_SASL_AUTH
#include <stdint.h>
#include <mempool.h>
#include <data-stack.h>
#include <failures.h>
#include <strfuncs.h>
#include <ioloop.h>
#include <lib-auth/auth-client.h>

void lib_init(void);
#endif

/* Application-specific. */

#include "smtpd.h"
#include "smtpd_sasl_glue.h"
#include "smtpd_chat.h"

#ifdef USE_SASL_AUTH

static struct ioloop *ioloop;

static void panic_handler(const char *fmt, va_list args) __attr_noreturn__;
static void fatal_handler(int unused_status, const char *fmt, va_list args)
    __attr_noreturn__;

static void panic_handler(const char *fmt, va_list args)
{
    msg_panic("%s", t_strdup_vprintf(fmt, args));
}

static void fatal_handler(int unused_status, const char *fmt, va_list args)
{
    msg_fatal("%s", t_strdup_vprintf(fmt, args));
}

static void error_handler(const char *fmt, va_list args)
{
    t_push();
    msg_error("%s", t_strdup_vprintf(fmt, args));
    t_pop();
}

static void warning_handler(const char *fmt, va_list args)
{
    t_push();
    msg_warn("%s", t_strdup_vprintf(fmt, args));
    t_pop();
}

static void info_handler(const char *fmt, va_list args)
{
    t_push();
    msg_info("%s", t_strdup_vprintf(fmt, args));
    t_pop();
}

/* smtpd_sasl_initialize - per-process initialization */
void    smtpd_sasl_initialize(void)
{
    lib_init();

    i_set_panic_handler(panic_handler);
    i_set_fatal_handler(fatal_handler);
    i_set_error_handler(error_handler);
    i_set_warning_handler(warning_handler);
    i_set_info_handler(info_handler);

    ioloop = io_loop_create(system_pool);
}

/* smtpd_sasl_connect - per-connection initialization */

void    smtpd_sasl_connect(SMTPD_STATE *state, const char *sasl_opts_name,
			   const char *sasl_opts_var)
{
    const struct auth_mech_desc *mech;
    unsigned i, mech_count;
    VSTRING *str;

    /*
     * Initialize SASL-specific state variables. Use long-lived storage for
     * base 64 conversion results, rather than local variables, to avoid
     * memory leaks when a read or write routine returns abnormally after
     * timeout or I/O error.
     */
    state->sasl_mechanism_list = 0;
    state->sasl_username = 0;
    state->sasl_method = 0;
    state->sasl_sender = 0;
    state->sasl_conn = 0;
    state->sasl_decoded = vstring_alloc(10);
    state->sasl_encoded = vstring_alloc(10);

    state->sasl_conn = auth_client_new_external(getpid(),
						"/etc/dovecot-auth",
						NULL, NULL);

    io_loop_set_running(ioloop);
    while (!auth_client_is_connected(state->sasl_conn))
	io_loop_handler_run(ioloop);

    str = vstring_alloc(20);
    mech = auth_client_get_available_mechs(state->sasl_conn, &mech_count);

    if (mech_count <= 0)
	msg_fatal("No SASL authentication mechanisms");

    for (i = 0; i < mech_count; i++) {
        if (VSTRING_LEN(str) > 0)
            VSTRING_ADDCH(str, ' ');
	vstring_strcat(str, mech[i].name);
    }
    state->sasl_mechanism_list = vstring_export(str);
}

/* smtpd_sasl_disconnect - per-connection cleanup */

void    smtpd_sasl_disconnect(SMTPD_STATE *state)
{
    if (state->sasl_mechanism_list) {
	myfree(state->sasl_mechanism_list);
	state->sasl_mechanism_list = 0;
    }
    if (state->sasl_conn) {
	auth_client_free(state->sasl_conn);
	state->sasl_conn = 0;
    }
    vstring_free(state->sasl_decoded);
    vstring_free(state->sasl_encoded);
}

static void auth_callback(struct auth_request *request,
                          struct auth_client_request_reply *reply,
                          const unsigned char *data, void *context)
{
    char   *myname = "smtpd_sasl_authenticate";
    SMTPD_STATE *state = context;

    if (reply == NULL) {
        state->sasl_state = -1;
        state->sasl_error = "501 Authentication process died";
        return;
    }

    if (data == NULL)
	data = "";

    switch (reply->result) {
    case AUTH_CLIENT_RESULT_CONTINUE:
	/*
	 * Send a server challenge. Avoid storing the challenge in a local
	 * variable, because we would leak memory when smtpd_chat_reply()
	 * does not return due to timeout or I/O error. sasl_encode64()
	 * null-terminates the result if the result buffer is large enough.
	 */
	if (msg_verbose)
	    msg_info("%s: uncoded challenge: %.*s",
		     myname, (int) reply->data_size, data);

        base64_encode(state->sasl_encoded,
                      (const char *) data + reply->reply_idx,
                      reply->data_size - reply->reply_idx);
	smtpd_chat_reply(state, "334 %s", vstring_str(state->sasl_encoded));

	/*
	 * Receive the client response. "*" means that the client gives up.
	 * XXX For now we ignore the fact that excessively long responses
	 * will be truncated. To handle such responses, we need to change
	 * smtpd_chat_query() so that it returns an error indication.
	 */
	smtpd_chat_query(state);
        if (strcmp(vstring_str(state->buffer), "*") == 0) {
            state->sasl_error = "501 Authentication aborted";
            break;
        }

        if (base64_decode(state->sasl_decoded,
                          vstring_str(state->buffer),
                          VSTRING_LEN(state->buffer)) == NULL) {
            state->sasl_error =
                "501 Authentication failed: malformed initial response";
            break;
        }

	if (msg_verbose)
	    msg_info("%s: decoded response: %.*s",
                     myname, (int) VSTRING_LEN(state->sasl_decoded),
                     vstring_str(state->sasl_decoded));

        auth_client_request_continue(request, vstring_str(state->sasl_decoded),
                                     VSTRING_LEN(state->sasl_decoded));
        break;

    case AUTH_CLIENT_RESULT_SUCCESS:
        state->sasl_username = mystrdup((const char *) data + reply->username_idx);
        state->sasl_state = 1;
        break;

    case AUTH_CLIENT_RESULT_FAILURE:
        /* see if we have error message */
        if (msg_verbose &&
            reply->data_size > 0 && data[reply->data_size-1] == '\0') {
            msg_info("%s: Authentication failed: ", myname,
                     (const char *) data);
        }
        state->sasl_state = -1;
        break;
    }

    if (state->sasl_error)
        state->sasl_state = -1;
}

/* smtpd_sasl_authenticate - per-session authentication */

char   *smtpd_sasl_authenticate(SMTPD_STATE *state,
				        const char *sasl_method,
				        const char *init_response)
{
    char   *myname = "smtpd_sasl_authenticate";
    struct auth_request *request;
    struct auth_request_info info;
    const char *error;

#define IFELSE(e1,e2,e3) ((e1) ? (e2) : (e3))

    if (msg_verbose)
	msg_info("%s: sasl_method %s%s%s", myname, sasl_method,
		 IFELSE(init_response, ", init_response ", ""),
		 IFELSE(init_response, init_response, ""));

    /*
     * Sanity check.
     */
    if (state->sasl_username || state->sasl_method)
	msg_panic("%s: already authenticated", myname);

    /*
     * SASL authentication protocol start-up. Process any initial client
     * response that was sent along in the AUTH command.
     */
    if (init_response) {
        if (base64_decode(state->sasl_decoded,
                          init_response, strlen(init_response)) == NULL)
	    return ("501 Authentication failed: malformed initial response");
	if (msg_verbose)
            msg_info("%s: decoded initial response %s", myname,
                     vstring_str(state->sasl_decoded));
    } else {
	VSTRING_RESET(state->sasl_decoded);
    }

    memset(&info, 0, sizeof(info));
    info.mech = sasl_method;
    info.protocol = var_smtpd_sasl_appname;
    info.flags = 0; /* FIXME: set TLS */
    /* FIXME: set local/remote IP */
    info.initial_resp_data = vstring_str(state->sasl_decoded);
    info.initial_resp_size = VSTRING_LEN(state->sasl_decoded);

    request = auth_client_request_new(state->sasl_conn, &info,
                                      auth_callback, state, &error);
    if (request == NULL) {
        if (msg_verbose)
            msg_info("%s: Authentication failed: %s", myname, error);
        return ("501 Authentication failed");
    }

    /*
     * Repeat until done or until the client gives up.
     */
    state->sasl_state = 0;
    state->sasl_error = NULL;
    io_loop_set_running(ioloop);
    while (state->sasl_state == 0)
	io_loop_handler_run(ioloop);

    /*
     * The authentication protocol was completed.
     */
    if (state->sasl_state < 0) {
        return (state->sasl_error ? state->sasl_error :
                "535 Error: authentication failed");
    }

    state->sasl_method = mystrdup(sasl_method);

    return (0);
}

/* smtpd_sasl_logout - clean up after smtpd_sasl_authenticate */

void    smtpd_sasl_logout(SMTPD_STATE *state)
{
    if (state->sasl_username) {
	myfree(state->sasl_username);
	state->sasl_username = 0;
    }
    if (state->sasl_method) {
	myfree(state->sasl_method);
	state->sasl_method = 0;
    }
}

#endif
